""" Mutant Phenotypes """
import json
from functools import reduce
from collections import defaultdict

from orangecontrib.bioinformatics.utils import serverfiles

DOMAIN = 'dictybase'
PHENOTYPES_FILENAME = 'mutant_phenotypes.json'


class DictyMutant:
    __slots__ = ('name', 'descriptor', 'genes', 'phenotypes', 'mutant_types')

    def __init__(self, mutant_entry: dict) -> None:
        """ A single Dictyostelium discoideum mutant from the Dictybase.

        :param mutant_entry: A single mutant entry from `curated mutants file
            <http://dictybase.org/db/cgi-bin/dictyBase/download/download.pl?area=mutant_phenotypes&ID=all-mutants.txt>`_.

        """
        # dictyBase ID
        self.name = mutant_entry.get('systematic_name', None)

        # dictyBase strain descriptor
        self.descriptor = mutant_entry.get('strain_descriptor', None)

        # all associated genes
        self.genes = mutant_entry.get('associated_genes', [])

        # all associated phenotypes
        self.phenotypes = mutant_entry.get('phenotypes', [])

        self.mutant_types = mutant_entry.get('mutant_types', None)

    def __repr__(self):
        return f'<dictyBase ID={self.name}, strain descriptor={self.descriptor}>'


class DictyMutants:
    DEFAULT_DATABASE_PATH = serverfiles.localpath(DOMAIN)  # use a default local folder for storing the genesets

    def __init__(self, file_path=None):
        """  A collection of Dictybase mutants as a dictionary of :obj:`DictyMutant` objects.
        """
        if file_path is None:
            file_path = serverfiles.localpath_download(DOMAIN, PHENOTYPES_FILENAME)

        with open(file_path, 'r') as fp:
            _mutants = [DictyMutant(mutant) for mutant in json.load(fp)]
            self._mutants = {m: m for m in _mutants}

    @classmethod
    def get_instance(cls):
        if not hasattr(cls, "_shared_dict"):
            dicty = DictyMutants()
            cls._shared_dict = dicty.__dict__
        instance = DictyMutants.__new__(DictyMutants)
        instance.__dict__ = cls._shared_dict
        return instance

    def mutants(self):
        return list(self._mutants.keys())

    def genes(self):
        return sorted(set(reduce(list.__add__, [self.mutant_genes(mutant) for mutant in self.mutants()], [])))

    def phenotypes(self):
        return sorted(set(reduce(list.__add__, [self.mutant_phenotypes(mutant) for mutant in self.mutants()], [])))

    def mutant_genes(self, mutant):
        return self._mutants[mutant].genes

    def mutant_phenotypes(self, mutant):
        return self._mutants[mutant].phenotypes

    def gene_mutants(self):
        dgm = defaultdict(set)
        for mutant, genes in [(mutant, self.mutant_genes(mutant)) for mutant in self.mutants()]:
            for gene in genes:
                dgm[gene].add(mutant)
        return dgm

    def phenotype_mutants(self):
        dpm = defaultdict(set)
        for mutant, phenotypes in [(mutant, self.mutant_phenotypes(mutant)) for mutant in self.mutants()]:
            for phenotype in phenotypes:
                dpm[phenotype].add(mutant)
        return dpm


def mutants():
    """ Return all :obj:`DictyMutant` objects.
    """
    return DictyMutants.get_instance().mutants()


def genes():
    """ Return a set of all genes referenced in the Dictybase.
    """
    return DictyMutants.get_instance().genes()


def phenotypes():
    """ Return a set of all phenotypes referenced in Dictybase.
    """
    return DictyMutants.get_instance().phenotypes()


def mutant_genes(mutant):
    """ Return a set of all genes referenced by a `mutant` in Dictybase.
    """
    return DictyMutants.get_instance().mutant_genes(mutant)


def mutant_phenotypes(mutant):
    """ Return a set of all phenotypes referenced by a `mutant` in Dictybase.
    """
    return DictyMutants.get_instance().mutant_phenotypes(mutant)


def gene_mutants():
    """ Return a dictionary { gene: set(mutant_objects for mutant), ... }.
    """
    return DictyMutants.get_instance().gene_mutants()


def phenotype_mutants():
    """ Return a dictionary { phenotype: set(mutant_objects for mutant), ... }.
    """
    return DictyMutants.get_instance().phenotype_mutants()


if __name__ == "__main__":
    dicty_mutants = mutants()
    mutant = list(dicty_mutants)[0]
    print(mutant.name, mutant.descriptor, mutant.genes)
    print(gene_mutants())



import unittest

import numpy as np

from Orange.data import Table, Domain, DiscreteVariable, ContinuousVariable
from Orange.clustering import DBSCAN, KMeans

from orangecontrib.bioinformatics.annotation.annotate_projection import (
    _angle,
    get_epsilon,
    assign_labels,
    _filter_clusters,
    labels_locations,
    annotate_projection,
    compute_concave_hulls,
    cluster_additional_points,
)


class TestAnnotateProjection(unittest.TestCase):
    def setUp(self):
        domain = Domain([ContinuousVariable("x"), ContinuousVariable("y")])
        data = np.array(
            [
                [1.0, 1.0],
                [1.0, 1.5],
                [1.5, 1.0],
                [1.2, 1.2],
                [1.7, 1.8],
                [10.0, 10.0],
                [10.0, 10.5],
                [10.5, 11.0],
                [10.5, 10.7],
                [10.5, 10.6],
                [10.5, 10.7],
                [10.5, 10.7],
            ]
        )
        self.data = Table(domain, data)

        domain_ann = Domain([ContinuousVariable("a"), ContinuousVariable("b"), ContinuousVariable("c")])
        data_ann = np.array(
            [
                [0.9, 1, 0.4],
                [0.9, 0.4, 0.2],
                [0.9, 0.4, 0.2],
                [0.9, 1, 0.2],
                [0.9, 0.4, 0.2],
                [0.2, 0.4, 0.7],
                [0.1, 0.4, 0.6],
                [0.2, 0.9, 0.2],
                [0.1, 0.4, 0.6],
                [0.1, 0.8, 0.6],
                [0.1, 0.8, 0.6],
                [0.1, 0.8, 0.6],
            ]
        )
        self.annotations = Table(domain_ann, data_ann)

    def test_annotate_projection(self):
        clusters, clusters_meta, eps = annotate_projection(
            self.annotations, self.data, clustering_algorithm=DBSCAN, labels_per_cluster=2, eps=3.0
        )

        self.assertListEqual(
            list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0])), ["C2"] * 5 + ["C1"] * 7
        )

        self.assertEqual(len(clusters_meta), 2)
        self.assertEqual(len(clusters_meta["C1"]), 3)
        self.assertEqual(len(clusters_meta["C2"]), 3)
        self.assertEqual(clusters_meta["C1"][0][0][0], 'b')
        self.assertAlmostEqual(clusters_meta["C1"][0][0][1], 4 / 7, 5)
        self.assertEqual(clusters_meta["C1"][0][1][0], 'c')
        self.assertAlmostEqual(clusters_meta["C1"][0][1][1], 3 / 7, 5)

        self.assertEqual(2, len(clusters_meta["C1"][1]))
        self.assertEqual(2, clusters_meta["C1"][2].shape[1])
        self.assertEqual(2, clusters_meta["C1"][2].shape[1])
        self.assertEqual(float, type(eps))

    def test_example_not_clustered(self):
        self.data[-1] = [23, 23]
        clusters, clusters_meta, eps = annotate_projection(
            self.annotations, self.data, clustering_algorithm=DBSCAN, eps=2, min_samples=3
        )

        self.assertListEqual(
            list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0])), ["C2"] * 5 + ["C1"] * 6 + ["?"]
        )

        self.assertEqual(len(clusters_meta), 2)
        self.assertEqual(len(clusters_meta["C1"]), 3)
        self.assertEqual(len(clusters_meta["C2"]), 3)
        self.assertEqual(clusters_meta["C1"][0][0][0], 'c')
        self.assertAlmostEqual(clusters_meta["C1"][0][0][1], 0.5, 5)
        self.assertEqual(clusters_meta["C1"][0][1][0], 'b')
        self.assertAlmostEqual(clusters_meta["C1"][0][1][1], 0.5, 5)

    def test_other_clustering(self):
        clusters, clusters_meta, locs = annotate_projection(
            self.annotations, self.data, clustering_algorithm=KMeans, n_clusters=2, labels_per_cluster=2
        )

        self.assertEqual(2, len(set(clusters.X[:, 0].flatten())))
        self.assertEqual(3, len(clusters_meta["C1"]))
        self.assertEqual(3, len(clusters_meta["C2"]))
        self.assertEqual(2, len(clusters_meta["C1"][0]))
        self.assertEqual(2, len(clusters_meta["C2"][0]))

    def test_labels_location(self):
        clusters = Table(Domain([DiscreteVariable("cl", values=["1", "2"])]), [[0]] * 6 + [[1]] * 6)
        locs = labels_locations(self.data, clusters)

        self.assertEqual(dict, type(locs))
        self.assertEqual(tuple, type(locs["1"]))
        self.assertEqual(tuple, type(locs["2"]))

    def test_get_epsilon(self):
        data = Table("iris")
        eps = get_epsilon(data)

        self.assertGreaterEqual(eps, 0.9)

    def test_compute_concave_hulls(self):
        data = Table("iris")[:, 2:4]
        clusters = Table(
            Domain([DiscreteVariable("cl", values=["1", "2", "3"])]), [[0]] * 50 + [[1]] * 50 + [[2]] * 50
        )
        hulls = compute_concave_hulls(data, clusters, epsilon=0.5)

        self.assertEqual(3, len(hulls))
        self.assertEqual(2, hulls["1"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["2"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["3"].shape[1])  # hull have x and y

    def test_compute_concave_hulls_subsampling(self):
        """
        When more than 1000 points passed they are sub-sampled in order to
        compute a concave hull
        """
        iris = Table("iris")
        data = Table(  # here we pass 1500 points
            Domain(iris.domain.attributes[2:4]), np.repeat(iris.X[:, 2:4], 10, axis=0)
        )
        clusters = Table(
            Domain([DiscreteVariable("cl", values=["1", "2", "3"])]),
            [[0]] * 50 * 10 + [[1]] * 50 * 10 + [[2]] * 50 * 10,
        )
        hulls = compute_concave_hulls(data, clusters, epsilon=0.5)

        self.assertEqual(3, len(hulls))
        self.assertEqual(2, hulls["1"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["2"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["3"].shape[1])  # hull have x and y

    def test_compute_concave_hulls_triangle(self):
        """
        Concave hull must also work for tree points - it is a special case
        """
        data = Table(Domain([ContinuousVariable("x"), ContinuousVariable("y")]), np.array([[1, 1], [1, 2], [2, 1]]))
        clusters = Table(Domain([DiscreteVariable("cl", values=["1"])]), [[0]] * 3)
        hulls = compute_concave_hulls(data, clusters, epsilon=0.5)

        self.assertEqual(1, len(hulls))
        self.assertEqual(2, hulls["1"].shape[1])  # hull have x and y

    def test_empty_annotations(self):
        ann = Table(Domain([]), np.empty((len(self.data), 0)))
        clusters, clusters_meta, eps = annotate_projection(ann, self.data)

        self.assertGreaterEqual(len(clusters), 0)
        self.assertEqual(len(clusters_meta), 0)

    def test_one_label(self):
        """
        Test whether having only one label works fine, in this case one cluster
        will not have a label assigned - this must work fine as well.
        """
        domain_ann = Domain([ContinuousVariable("a")])
        data_ann = np.array([[0.9], [0.9], [0.9], [0.9], [0.9], [0], [0], [0], [0], [0], [0], [0]])
        ann = Table(domain_ann, data_ann)
        clusters, clusters_meta, eps = annotate_projection(ann, self.data, eps=2)
        self.assertGreater(len(clusters), 0)
        self.assertGreater(len(clusters_meta), 0)
        self.assertGreater(len(clusters_meta["C1"][0]), 0)
        self.assertGreater(len(clusters_meta["C1"][1]), 0)
        self.assertGreater(len(clusters_meta["C1"][2]), 0)

    def test_socres_with_nan(self):
        self.annotations.X[0, 0] = np.nan
        self.annotations.X[1, 1] = np.nan
        self.annotations.X[1, 2] = np.nan

        self.annotations.X[2, 0] = np.nan
        self.annotations.X[2, 1] = np.nan
        self.annotations.X[2, 2] = np.nan

        clusters = Table(
            Domain([DiscreteVariable("Cluster", values=["C1", "C3"])]), np.array([[0] * 5 + [1] * 7]).reshape(-1, 1)
        )

        labels_dict, labels_items = assign_labels(clusters, self.annotations, 3)

        transformed_labels = list(map(labels_items.domain.attributes[0].repr_val, labels_items.X[:, 0]))
        self.assertListEqual(['b', 'a', '?', 'b', 'a', 'c', 'c', 'b', 'c', 'b', 'b', 'b'], transformed_labels)
        self.assertTupleEqual((0.4, 0.4), list(zip(*labels_dict["C1"]))[1])

        # check all nans
        self.annotations.X[:, :] = np.nan
        labels_dict, labels_items = assign_labels(clusters, self.annotations, 3)
        transformed_labels = list(map(labels_items.domain.attributes[0].repr_val, labels_items.X[:, 0]))
        self.assertListEqual(['?'] * len(self.annotations), transformed_labels)
        self.assertFalse("C1" in labels_dict)
        self.assertFalse("C3" in labels_dict)

    def test_cluster_additional_points_simple(self):
        """
        Test with the simple convex hull
        """
        hull = {"C1": np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.0]])}
        x, y = np.meshgrid(np.arange(-1.0, 2.0, 0.2), np.arange(-1, 2.0, 0.2))
        points = Table(
            Domain([ContinuousVariable("x"), ContinuousVariable("y")]),
            np.concatenate((x.reshape(-1, 1), y.reshape(-1, 1)), axis=1),
        )

        clusters = cluster_additional_points(points, hull)
        expected = [0.0 if 0 < x < 1 and 0 < y < 1 else np.nan for x, y in points.X]
        np.testing.assert_array_equal(clusters.X.flatten(), expected)

    def test_cluster_additional_points_concave(self):
        """
        Test with the concave hull
        """
        hull = {"C1": np.array([[0.0, 0.0], [0.3, 0.5], [0.0, 1.0], [1.0, 1.0], [0.7, 0.5], [1.0, 0.0]])}
        p = [
            [-1.0, -1],  # point that is totally out polygon
            [0.2, 0.5],  # point outside the polygon, but close
            [0.8, 0.5],  # point outside the polygon, but close
            [0.5, 0.5],  # point in
            [0.5, 0.7],  # point in
            [0.5, 0.3],  # point in
            [0.2, 0.9],  # point in
            [0.2, 0.1],  # point in
        ]
        points = Table(Domain([ContinuousVariable("x"), ContinuousVariable("y")]), p)

        clusters = cluster_additional_points(points, hull)
        expected = [np.nan] * 3 + [0.0] * 5
        np.testing.assert_array_equal(clusters.X.flatten(), expected)

    points = [
        [-1.0, -1],  # point that is totally out  of any polygon
        [0.2, 0.5],  # point outside, but in concave part of C1
        [0.8, 0.5],  # point outside, but in concave part of C1
        [0.5, 0.5],  # point in C1
        [0.5, 0.7],  # point in C1
        [0.5, 0.3],  # point in C1
        [0.2, 0.9],  # point in C1
        [0.2, 0.1],  # point in C1
        [-0.5, 0.1],  # point in C2
        [-0.8, 0.1],  # point in C2
        [-0.5, 0.8],  # point out, but in concave part of C2
        [-0.6, 0.9],  # point out, but in concave part of C2
    ]

    hull = {
        "C1": np.array([[0.0, 0.0], [0.3, 0.5], [0.0, 1.0], [1.0, 1.0], [0.7, 0.5], [1.0, 0.0]]),
        "C2": np.array([[0.0, 0.0], [0.0, 1.0], [-0.5, 0.7], [-1.0, 1.0], [-1.0, 0.0]]),
    }

    def test_cluster_additional_points_more_clusters(self):
        """
        Test with more concave hulls
        """

        points_table = Table(Domain([ContinuousVariable("x"), ContinuousVariable("y")]), self.points)

        clusters = cluster_additional_points(points_table, self.hull)
        clusters = list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
        expected = ["?"] * 3 + ["C1"] * 5 + ["C2"] * 2 + ["?"] * 2
        np.testing.assert_array_equal(clusters, expected)

    def test_cluster_additional_points_with_existing_domain(self):
        """
        Test with the concave hull
        """
        points = Table(Domain([ContinuousVariable("x"), ContinuousVariable("y")]), self.points)
        attr = DiscreteVariable("Clusters", values=["C2", "C1"])

        clusters = cluster_additional_points(points, self.hull, cluster_attribute=attr)
        cluster_map = list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
        expected = ["?"] * 3 + ["C1"] * 5 + ["C2"] * 2 + ["?"] * 2
        np.testing.assert_array_equal(cluster_map, expected)
        self.assertEqual(attr, clusters.domain.attributes[0])
        self.assertListEqual(attr.values, clusters.domain.attributes[0].values)

    def test_angle(self):
        # test cases with first vector from x, y = [1, 0] to x, y = [0, 0]
        # and second from x, y = [0, 0] to points in test_cases
        v1 = ([1, 0], [0, 0])
        v2_1 = [0, 0]
        test_cases = [[1, 0], [1, -1], [0, -1], [-1, -1], [-1, 0], [-1, 1], [0, 1], [1, 1]]
        real_angles = [np.pi * i / 4 for i in range(8)]
        computed_angles = [_angle(v1, (v2_1, x)) for x in test_cases]
        np.testing.assert_array_almost_equal(computed_angles, real_angles)

        # test cases with first vector from x, y = [2, 2] to x, y = [1, 2]
        # and second from x, y = [1, 2] to points in test_cases
        v1 = ([2, 2], [1, 2])
        v2_1 = np.array([1, 2])
        test_cases = np.array([[1, 0], [1, -1], [0, -1], [-1, -1], [-1, 0], [-1, 1], [0, 1], [1, 1]]) + v2_1
        real_angles = [np.pi * i / 4 for i in range(8)]
        computed_angles = [_angle(v1, (v2_1, x)) for x in test_cases]
        np.testing.assert_array_almost_equal(computed_angles, real_angles)

        # some more complex cases
        self.assertAlmostEqual(_angle(([-2, 1], [0, 0]), ([0, 0], [1, 2])), np.pi / 2)
        self.assertAlmostEqual(_angle(([-4, 1], [-2, 0]), ([-2, 0], [-1, 2])), np.pi / 2)
        self.assertAlmostEqual(_angle(([-4, 3], [-2, 2]), ([-2, 2], [-1, 4])), np.pi / 2)

    def test_cluster_filtering(self):
        """
        Cluster filtering was introduced in order to remove clusters that
        does not have any labels.
        """

        def assert_dict_same(d1, d2):
            self.assertTrue(len(d1) == len(d2) and sorted(d1) == sorted(d2))

        clusters = Table(Domain([DiscreteVariable("Clusters", values=["C1", "C2"])]), [[0]] * 6 + [[1]] * 5)
        metas = {"C1": "test1", "C2": "test 2"}

        new_clusters, new_metas = _filter_clusters(clusters, metas)

        # check old table unchanged
        self.assertEqual(11, len(clusters))
        self.assertListEqual(["C1", "C2"], clusters.domain["Clusters"].values)
        np.testing.assert_array_equal(np.array([[0]] * 6 + [[1]] * 5), clusters.X)
        assert_dict_same(metas, {"C1": "test1", "C2": "test 2"})

        # new clusters should be same in this case
        self.assertEqual(11, len(new_clusters))
        self.assertListEqual(["C1", "C2"], new_clusters.domain["Clusters"].values)
        np.testing.assert_array_equal(np.array([[0]] * 6 + [[1]] * 5), new_clusters.X)
        assert_dict_same(new_metas, {"C1": "test1", "C2": "test 2"})

        clusters = Table(Domain([DiscreteVariable("Clusters", values=["C1", "C2"])]), [[0]] * 6 + [[1]] * 5)
        metas = {"C1": "test1"}

        new_clusters, new_metas = _filter_clusters(clusters, metas)

        # check old table unchanged
        self.assertEqual(11, len(clusters))
        self.assertListEqual(["C1", "C2"], clusters.domain["Clusters"].values)
        np.testing.assert_array_equal(np.array([[0]] * 6 + [[1]] * 5), clusters.X)
        assert_dict_same(metas, {"C1": "test1"})

        # new clusters should be same in this case
        self.assertEqual(11, len(new_clusters))
        self.assertListEqual(["C1"], new_clusters.domain["Clusters"].values)
        np.testing.assert_array_equal(np.array([[0]] * 6 + [[np.nan]] * 5), new_clusters.X)
        assert_dict_same(new_metas, {"C1": "test1"})



""" Orange data table and other data manipulation utilities """
import numbers
from types import SimpleNamespace
from typing import Tuple, Sequence

import Orange.data


class TableAnnotation(SimpleNamespace):
    """ Data Table hints """

    # Organism in data table
    tax_id: str = 'taxonomy_id'

    # This indicates position of genes in data table
    gene_as_attr_name: str = 'gene_as_attribute_name'

    # This indicates a column name (if genes are in rows)
    gene_id_column: str = 'gene_id_column'

    # This indicates attribute name (if genes are in columns)
    gene_id_attribute: str = 'gene_id_attribute'


# species
TAX_ID = 'taxonomy_id'

# Will be set to True if gene names are represented as attribute names.
# If gene names are in rows, we set this value to False. (user must select proper column index)
GENE_AS_ATTRIBUTE_NAME = 'gene_as_attribute_name'

# Name of the column where rows are gene ids
GENE_ID_COLUMN = 'gene_id_column'

# Name of the variable attribute that holds gene id
GENE_ID_ATTRIBUTE = 'gene_id_attribute'

# Error strings
ERROR_ON_MISSING_ANNOTATION = 'Missing annotation on gene IDs and organism in the input data.'
ERROR_ON_MISSING_GENE_ID = 'Missing gene ID information in the input data'
ERROR_ON_MISSING_TAX_ID = 'Missing organism information in the input data'

col_spec = Sequence[Tuple[Orange.data.Variable, Sequence[numbers.Real]]]
metas_spec = Sequence[Tuple[Orange.data.Variable, Sequence[str]]]



from Orange.widgets import settings


class SetContextHandler(settings.ContextHandler):
    def __init__(self, match_imperfect=False):
        super().__init__()
        self.match_imperfect = match_imperfect

    def match(self, context, items):
        items = set(items)

        if self.match_imperfect:
            intersection, union = items & context.items, items | context.items
            if len(union) > 0:
                return len(intersection) / len(union)
            else:
                return 0
        else:
            return 2 if items == context.items else 0

    def new_context(self, items):
        ctx = super().new_context()
        ctx.items = frozenset(items)
        return ctx

    def settings_to_widget(self, widget, *args):
        super().settings_to_widget(widget, *args)

        context = widget.current_context
        if context is None:
            return

        for setting, data, instance in self.provider.traverse_settings(context.values, widget):
            if isinstance(setting, settings.ContextSetting) and setting.name in data:
                value = self.decode_setting(settings, data[setting.name])
                setattr(instance, setting.name, value)


class OrganismContextHandler(settings.ContextHandler):
    def __init__(self):
        super().__init__()

    def match(self, context, taxonomy_id, *args):
        if not context.organism == taxonomy_id:
            return self.NO_MATCH

        return self.PERFECT_MATCH

    def new_context(self, tax_id):
        context = super().new_context()
        context.organism = tax_id
        return context

    def settings_from_widget(self, widget, *args):
        super().settings_from_widget(widget, *args)

        context = widget.current_context
        if context is None:
            return

        # get taxonomy id from the widget
        context.organism = widget.tax_id



import numpy as np
import pandas as pd
from cellannotation.annotate_samples import PFUN_BINOMIAL, SCORING_EXP_RATIO, AnnotateSamples

from Orange.data import Table, Domain, ContinuousVariable

from orangecontrib.bioinformatics.ncbi.gene import GeneInfo
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID


class ScoringNotImplemented(Exception):
    pass


class AnnotateSamplesMeta:
    """
    AnnotateSamples is a meta class used for the annotation of data items with
    the labels Mann-Whitney U test for selecting important values and
    the Hyper-geometric for assigning the labels. This class is a proxy to the
    external library - cell-annotation.

    Example for full annotation:

    >>> from Orange.data import Table
    >>> from orangecontrib.bioinformatics.utils import serverfiles
    >>> from orangecontrib.bioinformatics.annotation.annotate_samples import \
    ...     AnnotateSamples
    >>> from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID
    >>> from Orange.data.filter import FilterString, Values
    >>>
    >>> data = Table("https://datasets.orange.biolab.si/sc/aml-1k.tab.gz")
    >>> data.attributes[TAX_ID] = "9606"  # table needs to have an organism ID
    >>> markers_path = serverfiles.localpath_download(
    ...     'marker_genes','panglao_gene_markers.tab')
    >>> marker = Table(markers_path)
    >>>
    >>> # filter only human markers
    >>> f = FilterString("Organism", FilterString.Equal, "Human")
    >>> markers = Values([f])(marker)
    >>>
    >>> annotator = AnnotateSamples
    >>> z = annotator.mann_whitney_test(data)
    >>> scores, p_val = AnnotateSamples.assign_annotations(z, markers, data)
    >>> scores = annotator.filter_annotations(scores, p_val, p_threshold=0.05)
    """

    @staticmethod
    def _to_pandas(data, use_entrez_id=False):
        """
        Transform Orange table to pandas dataframe.

        Parameters
        ----------
        data : Orange.data.Table
            Orange table with gene data. We suppose that genes have `Entrez ID`
            assigned.
        use_entrez_id : bool
            This parameter tells whether to use Entrez ID as column names.
            When it is False variables names will be used

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame with EntrezID
        list
            List of booleans with True on location of included columns in
            dataframe.
        """
        if use_entrez_id:
            entrez_ids = [x.attributes.get("Entrez ID") for x in data.domain.attributes]
            has_entrez_id = np.array([x is not None for x in entrez_ids])
            return (
                pd.DataFrame(data.X[:, has_entrez_id], columns=map(str, np.array(entrez_ids)[has_entrez_id])),
                has_entrez_id.tolist(),
            )
        else:
            return pd.DataFrame(data.X, columns=map(str, data.domain.attributes)), None

    @staticmethod
    def mann_whitney_test(data):
        """
        Compute z values with Mann-Whitney U test.

        Parameters
        ----------
        data : Orange.data.Table
            Tabular data with gene expressions

        Returns
        -------
        Orange.data.Table
            Z-value for each item.
        """
        df, columns = AnnotateSamplesMeta._to_pandas(data, use_entrez_id=False)
        df_z_values = AnnotateSamples.mann_whitney_test(df)

        z_table = Table(data.domain, df_z_values.values, data.Y, metas=data.metas)

        return z_table

    @staticmethod
    def assign_annotations(
        z_values, available_annotations, data, z_threshold=1, p_value_fun=PFUN_BINOMIAL, scoring=SCORING_EXP_RATIO
    ):
        """
        The function gets a set of attributes (e.g. genes) for each cell and
        attributes for each annotation. It returns the annotations significant
        for each cell.

        Parameters
        ----------
        z_values : Orange.data.Table
            Table which show z values for each item
        available_annotations : Orange.data.Table
            Available annotations (e.g. cell types)
        z_threshold : float
            The threshold for selecting the attribute. For each item the
            attributes with z-value above this value are selected.
        p_value_fun : str, optional (defaults: TEST_BINOMIAL)
            A function that calculates p-value. It can be either
            PFUN_BINOMIAL that uses statistics.Binomial().p_value or
            PFUN_HYPERGEOMETRIC that uses hypergeom.sf.
        data : Orange.data.Table
            Tabular data with gene expressions - we need that to compute scores.
        scoring : str, optional (default=SCORING_EXP_RATIO)
            Type of scoring

        Returns
        -------
        Orange.data.Table
            Annotation probabilities
        Orange.data.Table
            Annotation fdrs
        """
        # checks that assures that data are ok
        assert TAX_ID in data.attributes, "The input table needs to have a " "tax_id attribute"
        assert any(
            "Entrez ID" in x.attributes for x in data.domain.attributes
        ), "Input data do not contain gene expression data."

        # retrieve number of genes
        tax_id = data.attributes[TAX_ID]
        n = len(GeneInfo(tax_id))  # number of genes for organism

        # transform data to pandas dataframe
        df_z_values, _ = AnnotateSamplesMeta._to_pandas(z_values, use_entrez_id=True)
        df_data, _ = AnnotateSamplesMeta._to_pandas(data, use_entrez_id=True)
        # transform marker genes
        columns = list(map(str, available_annotations.domain.metas))
        # the framework recognizes Gene instead of Entrez ID
        columns[columns.index("Entrez ID")] = "Gene"
        df_available_annotations = pd.DataFrame(available_annotations.metas, columns=columns)
        df_available_annotations = df_available_annotations[df_available_annotations["Gene"] != "?"]

        # call the method
        scores, fdrs = AnnotateSamples.assign_annotations(
            df_z_values,
            df_available_annotations,
            df_data,
            n,
            z_threshold=z_threshold,
            p_value_fun=p_value_fun,
            scoring=scoring,
        )

        # create orange tables
        domain = Domain([ContinuousVariable(ct) for ct in scores.columns.values])
        scores_table = Table(domain, scores.values)
        fdrs_table = Table(domain, fdrs.values)

        return scores_table, fdrs_table

    @staticmethod
    def filter_annotations(scores, p_values, return_nonzero_annotations=True, p_threshold=0.05):
        """
        This function filters the probabilities on places that do not reach the
        threshold for p-value and filter zero columns
        return_nonzero_annotations is True.

        Parameters
        ----------
        scores : Orange.data.Table
            Scores for each annotations for each cell
        p_values : Orange.data.Table
            p-value scores for annotations for each cell
        return_nonzero_annotations : bool
            Flag that enables filtering the non-zero columns.
        p_threshold : float
            A threshold for accepting the annotations. Annotations that has FDR
            value bellow this threshold are used.

        Returns
        -------
        Orange.data.Table
            Filtered scores for each annotations for each cell
        """
        df_scores, _ = AnnotateSamplesMeta._to_pandas(scores)
        df_p_values, _ = AnnotateSamplesMeta._to_pandas(p_values)
        scores = AnnotateSamples.filter_annotations(df_scores, df_p_values, return_nonzero_annotations, p_threshold)

        domain = Domain([ContinuousVariable(ct) for ct in scores.columns.values])
        scores_table = Table(domain, scores.values)
        return scores_table



import unittest

from orangecontrib.bioinformatics.ncbi import taxonomy


class TestTaxonomy(unittest.TestCase):

    human = '9606'
    dicty = '44689'
    dog = '9615'

    def setUp(self) -> None:
        self.tax_obj = taxonomy.Taxonomy()
        self.assertGreater(len(self.tax_obj.taxids()), 1500000)

    def test_common_taxonomy(self):
        self.assertGreater(len(taxonomy.common_taxids()), 0)

        self.assertEqual(taxonomy.name(self.human), 'Homo sapiens')
        self.assertEqual(taxonomy.name(self.dicty), 'Dictyostelium discoideum')

        self.assertEqual(taxonomy.species_name_to_taxid('Homo sapiens'), self.human)
        self.assertEqual(taxonomy.species_name_to_taxid('Dictyostelium discoideum'), self.dicty)

        self.assertGreater(len(taxonomy.shortname(self.human)), 0)
        self.assertGreater(len(taxonomy.shortname(self.dicty)), 0)

    def test_uncommon_taxonomy(self):
        self.assertTrue(self.dog not in taxonomy.common_taxids())
        self.assertEqual(taxonomy.name(self.dog), 'Canis lupus familiaris')

        # not supported yet.
        self.assertIsNone(taxonomy.species_name_to_taxid('Canis lupus familiaris'))
        self.assertFalse(len(taxonomy.shortname(self.dog)))

    def test_human(self):
        self.assertTrue(isinstance(self.tax_obj.get_entry(self.human), taxonomy.utils.Taxon))

        self.assertRaises(taxonomy.utils.UnknownSpeciesIdentifier, self.tax_obj.get_entry, 'unknown_tax')

        self.assertTrue(('man', 'common name') in self.tax_obj.other_names(self.human))
        self.assertEqual(self.tax_obj.rank(self.human), 'species')
        self.assertEqual(self.tax_obj.parent(self.human), '9605')

        self.assertGreater(len(self.tax_obj.search('Homo sapiens', exact=True)), 0)
        self.assertGreater(len(self.tax_obj.lineage(self.human)), 0)
        self.assertGreater(len(self.tax_obj.get_all_strains(self.human)), 0)

        subnodes = self.tax_obj.subnodes(self.human)
        self.assertTrue(len(subnodes) == 2)
        self.assertTrue('63221' in subnodes)
        self.assertTrue('741158' in subnodes)

        neanderthal = self.tax_obj.get_entry('63221')
        self.assertTrue(neanderthal.parent_tax_id == self.tax_obj.get_species('63221'))
        self.assertEqual(neanderthal.name, 'Homo sapiens neanderthalensis')

        denisovan = self.tax_obj.get_entry('741158')
        self.assertTrue(denisovan.parent_tax_id == self.tax_obj.get_species('741158'))
        self.assertEqual(denisovan.name, "Homo sapiens subsp. 'Denisova'")

        self.assertTrue(len(taxonomy.search('Homo sapiens', exact=True)) == 1)
        self.assertIn(self.human, taxonomy.search('Homo sapiens', exact=True))

        linage = taxonomy.lineage(self.human)
        self.assertEqual(linage[0], '1')
        self.assertEqual(linage[-1], self.tax_obj.parent(self.human))

        search_result = taxonomy.search('Homo sapiens')
        self.assertTrue(len(search_result))
        self.assertIn(self.human, search_result)

        # unclassified Mammalia: Homo sapiens x Mus musculus hybrid cell line
        self.assertIn('1131344', search_result)

        # subnodes are included
        self.assertIn(neanderthal.tax_id, taxonomy.search('Homo sapiens', only_species=False))
        self.assertIn(denisovan.tax_id, taxonomy.search('Homo sapiens', only_species=False))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import setuptools

if __name__ == '__main__':
    setuptools.setup(
        use_scm_version=True,
        setup_requires=[
            'setuptools-scm',
            'setuptools>=40.0'
        ],
        install_requires=[
            'serverfiles==0.3.0',
            'requests==2.22.0',
            'requests-cache==0.5.2',
            'genesis-pyapi==1.2.1',
            'pyclipper==1.1.0.post1',
            'Orange3>=3.22.0',
            'cellannotation',
            # Versions are determined by Orange
            'scipy',
            'numpy',
        ],
        extras_require={
            'doc': ['sphinx', 'recommonmark'],
            'test': [
                'flake8~=3.7.8',
                'flake8-comprehensions~=2.2.0',
                'flake8-black~=0.1.0',
                'pep8-naming~=0.8.2',
                'isort~=4.3.21',
                'pre-commit~=1.18.3',
                'pytest~=5.1.0',
                'pytest-cov~=2.7.1',
                'coverage~=4.5.4',
                'codecov~=2.0.15'
            ],
        },
    )



import unittest
from os.path import basename, normpath

from Orange.data import Table

from orangecontrib.bioinformatics.ncbi.gene import ENTREZ_ID, Gene, GeneInfo, GeneMatcher


class TestGene(unittest.TestCase):
    def test_load_attributes(self):
        g = Gene()
        g.load_attributes(('Human', '9606', '920', 'CD4'), attributes=('species', 'tax_id', 'gene_id', 'symbol'))
        self.assertEqual(g.species, 'Human')
        self.assertEqual(g.tax_id, '9606')
        self.assertEqual(g.gene_id, '920')
        self.assertEqual(g.symbol, 'CD4')
        self.assertEqual(str(g), '<Gene symbol=CD4, tax_id=9606, gene_id=920>')

        self.assertIsNone(g.input_identifier)
        self.assertIsNone(g.synonyms)

    def test_homologs(self):
        gm = GeneMatcher('9606')
        gm.genes = ['920']
        g = gm.genes[0]

        self.assertIsNotNone(g.homologs)
        self.assertTrue(len(g.homologs))
        self.assertIn('10090', g.homologs)
        self.assertEqual(g.homology_group_id, '513')

        self.assertEqual(g.homolog_gene('10090'), '12504')
        self.assertIsNone(g.homolog_gene('Unknown_taxonomy'))


class TestGeneMatcher(unittest.TestCase):
    def test_synonym_multiple_matches(self):
        gm = GeneMatcher('9606')
        gm.genes = ['HB1']
        gene = gm.genes[0]
        self.assertEqual(gene.input_identifier, 'HB1')
        # Gene matcher should not find any unique match
        self.assertEqual(gene.gene_id, None)

    def test_symbol_match_scenario(self):
        gm = GeneMatcher('9606')
        gm.genes = ['SCN5A']
        gene = gm.genes[0]

        self.assertEqual(gene.input_identifier, 'SCN5A')
        self.assertEqual(gene.symbol, 'SCN5A')
        self.assertEqual(gene.gene_id, '6331')

    def test_different_input_identifier_types(self):
        gm = GeneMatcher('9606')
        gm.genes = ['CD4', '614535', 'HB-1Y', 'ENSG00000205426']

        for gene in gm.genes:
            self.assertIsNotNone(gene.description)
            self.assertIsNotNone(gene.tax_id)
            self.assertIsNotNone(gene.species)
            self.assertIsNotNone(gene.gene_id)

    def test_taxonomy_change(self):
        gm = GeneMatcher('4932')
        self.assertEqual(gm.tax_id, '4932')
        self.assertEqual(basename(normpath(gm.gene_db_path)), '4932.sqlite')

        gm.tax_id = '9606'
        self.assertEqual(gm.tax_id, '9606')
        self.assertEqual(basename(normpath(gm.gene_db_path)), '9606.sqlite')

    def test_match_table_column(self):
        gm = GeneMatcher('4932')

        data = gm.match_table_column(Table('brown-selected.tab'), 'gene')
        self.assertTrue(ENTREZ_ID in data.domain)

    def test_match_table_attributes(self):
        gm = GeneMatcher('4932')

        data = Table('brown-selected.tab')
        data = Table.transpose(data, feature_names_column='gene')
        gm.match_table_attributes(data)

        for column in data.domain.attributes:
            self.assertTrue(ENTREZ_ID in column.attributes)


class TestGeneInfo(unittest.TestCase):
    def test_gene_info(self):
        gi = GeneInfo('9606')
        gene = gi['6331']

        self.assertTrue('6331' in gi)
        self.assertIsInstance(gene, Gene)
        self.assertIsNotNone(gene.description)
        self.assertIsNotNone(gene.tax_id)
        self.assertIsNotNone(gene.species)
        self.assertIsNotNone(gene.gene_id)
        # must be None
        self.assertIsNone(gene.input_identifier)


if __name__ == '__main__':
    unittest.main()



""" NCBI GeneInformation module """
import json
import sqlite3
import contextlib
from typing import Dict, List, Tuple, Optional

from Orange.data import Table, Domain, StringVariable

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.ncbi.taxonomy import species_name_to_taxid
from orangecontrib.bioinformatics.ncbi.gene.config import DOMAIN, ENTREZ_ID, query, query_exact, gene_info_attributes
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation


class Gene:
    """ Representation of gene summary. """

    __slots__ = gene_info_attributes + ('input_identifier',)

    def __init__(self, input_identifier: Optional[str] = None):
        """
        If we want to match gene to it's corresponding Entrez ID we must,
        upon class initialization, provide some `input identifier`. This way
        :class:`GeneMatcher` will know what to match it against in Gene Database.

        Parameters
        ----------
        input_identifier : str
            This can be any of the following: symbol, synonym, locus tag, other database id, ...
        """
        self.input_identifier = input_identifier

    def __getattr__(self, attribute):
        if attribute not in self.__slots__:
            return None

    def __repr__(self):
        return f'<Gene symbol={self.symbol}, tax_id={self.tax_id}, gene_id={self.gene_id}>'

    def load_attributes(self, values: Tuple[str, ...], attributes: Tuple[str, ...] = gene_info_attributes):
        for attr, val in zip(attributes, values):
            setattr(self, attr, json.loads(val) if attr in ('synonyms', 'db_refs', 'homologs') else val)

    def homolog_gene(self, taxonomy_id: str) -> Optional[str]:
        """ Returns gene homolog for given organism.

        Parameters
        ----------
        taxonomy_id: str
            Taxonomy id of target organism.

        Returns
        -------
        str
            Entrez ID (if available).
        """
        return self.homologs.get(taxonomy_id, None)


class GeneMatcher:
    """ Gene name matching interface. """

    def __init__(self, tax_id: str, progress_callback=None, auto_start=True):
        """

        Parameters
        ----------
        tax_id:: str
            Taxonomy id of target organism.

        """
        self._tax_id: str = tax_id
        self._genes: List[Gene] = []
        self._progress_callback = progress_callback
        self._auto_start = auto_start
        self.gene_db_path = self._gene_db_path()

    @property
    def tax_id(self):
        return self._tax_id

    @tax_id.setter
    def tax_id(self, tax_id: str) -> None:
        self._tax_id = tax_id
        self.gene_db_path = self._gene_db_path()

    @property
    def genes(self) -> List[Gene]:
        return self._genes

    @genes.setter
    def genes(self, genes: List[str]) -> None:
        self._genes = [Gene(input_identifier=gene) for gene in genes]
        if self._auto_start:
            self._match()

    def get_known_genes(self) -> List[Gene]:
        """ Return Genes with known Entrez ID

        Returns
        -------
        :class:`list` of :class:`Gene` instances
            Genes with unique match

        """
        return [gene for gene in self.genes if gene.gene_id]

    def to_data_table(self, selected_genes: Optional[List[str]] = None) -> Table:
        """ Transform GeneMatcher results to Orange data table.

        Optionally we can provide a list of genes (Entrez Ids).
        The table on the output will be populated only with provided genes.

        Parameters
        ----------
        selected_genes: list
            List of Entrez Ids

        Returns
        -------
        Orange.data.Table
            Summary of Gene info in tabular format
        """
        data_x = []
        metas = [
            StringVariable('Input gene ID'),
            StringVariable(ENTREZ_ID),
            StringVariable('Symbol'),
            StringVariable('Synonyms'),
            StringVariable('Description'),
            StringVariable('Other IDs'),
            StringVariable('Type of gene'),
            StringVariable('Chromosome'),
            StringVariable('Map location'),
            StringVariable('Locus tag'),
            StringVariable('Symbol from nomenclature authority'),
            StringVariable('Full name from nomenclature authority'),
            StringVariable('Nomenclature status'),
            StringVariable('Other designations'),
            StringVariable('Species'),
            StringVariable('Taxonomy ID'),
        ]
        domain = Domain([], metas=metas)

        genes: List[Gene] = self.genes
        if selected_genes is not None:
            genes = [gene for gene in self.genes if str(gene.gene_id) in selected_genes]

        for gene in genes:
            db_refs = (
                ', '.join('{}: {}'.format(key, val) for (key, val) in gene.db_refs.items()) if gene.db_refs else ''
            )
            synonyms = ', '.join(gene.synonyms) if gene.synonyms else ''

            line = [
                gene.input_identifier,
                gene.gene_id,
                gene.symbol,
                synonyms,
                gene.description,
                db_refs,
                gene.type_of_gene,
                gene.chromosome,
                gene.map_location,
                gene.locus_tag,
                gene.symbol_from_nomenclature_authority,
                gene.full_name_from_nomenclature_authority,
                gene.nomenclature_status,
                gene.other_designations,
                species_name_to_taxid(gene.species),
                gene.tax_id,
            ]

            data_x.append(line)

        table = Table(domain, data_x)
        table.name = 'Gene Matcher Results'
        table.attributes[TableAnnotation.tax_id] = self.tax_id
        table.attributes[TableAnnotation.gene_as_attr_name] = False
        table.attributes[TableAnnotation.gene_id_column] = ENTREZ_ID
        return table

    def match_table_column(
        self, data_table: Table, column_name: str, target_column: Optional[StringVariable] = None
    ) -> Table:
        """ Helper function for gene name matching with :class:`Orange.data.Table`.

        Give a column of genes, GeneMatcher will try to map genes to their
        corresponding Entrez Ids.


        Parameters
        ----------
        data_table: :class:`Orange.data.Table`
            Data table

        column_name: str
            Name of the column where gene symbols are located

        target_column: :class:`StringVariable`
            Column where we store Entrez Ids.
            Defaults to StringVariable(ncbi.gene.config.NCBI_ID)

        Returns
        -------
        :class:`Orange.data.Table`
            Data table with a column of Gene Ids
        """

        if column_name in data_table.domain:
            self.genes = data_table.get_column_view(column_name)[0]

            if target_column is None:
                target_column = StringVariable(ENTREZ_ID)

            domain = Domain([], metas=[target_column])
            data = [[str(gene.gene_id) if gene.gene_id else '?'] for gene in self.genes]
            table = Table(domain, data)

            return Table.concatenate([data_table, table])

    def match_table_attributes(self, data_table):
        """ Helper function for gene name matching with :class:`Orange.data.Table`.

        Match table attributes and if a unique match is found create a new column attribute
        for Entrez Id. Attribute name is defined here: `orangecontrib.bioinformatics.ncbi.gene.config.NCBI_ID`


        Parameters
        ----------
        data_table: :class:`Orange.data.Table`
            Data table

        Returns
        -------
        :class:`Orange.data.Table`
            Data table column attributes are populated with Entrez Ids

        """
        input_gene_names = [var.name for var in data_table.domain.attributes]

        if input_gene_names:
            self.genes = input_gene_names

            for gene in self.genes:
                if gene.gene_id:
                    data_table.domain[gene.input_identifier].attributes[ENTREZ_ID] = gene.gene_id

    def match_genes(self):
        self._match()

    def _gene_db_path(self):
        return serverfiles.localpath_download(DOMAIN, f'{self.tax_id}.sqlite')

    def _match(self):
        synonyms, db_refs = 4, 5

        with contextlib.closing(sqlite3.connect(self.gene_db_path)) as con:
            with con as cursor:
                for gene in self.genes:

                    if self._progress_callback:
                        self._progress_callback()

                    search_param = gene.input_identifier.lower()

                    if search_param:
                        match_statement = (
                            '{gene_id symbol locus_tag symbol_from_nomenclature_authority}:^"' + search_param + '"'
                        )
                        match = cursor.execute(query_exact, (match_statement,) + tuple([search_param] * 4)).fetchall()
                        # if unique match
                        if len(match) == 1:
                            gene.load_attributes(match[0])
                            continue

                        match = cursor.execute(query, (f'synonyms:"{search_param}"',)).fetchall()
                        synonym_matched_rows = [
                            m for m in match if search_param in (x.lower() for x in json.loads(m[synonyms]))
                        ]
                        # if unique match
                        if len(synonym_matched_rows) == 1:
                            gene.load_attributes(synonym_matched_rows[0])
                            continue

                        match = cursor.execute(query, (f'db_refs:"{search_param}"',)).fetchall()
                        db_ref_matched_rows = [
                            m for m in match if search_param in (x.lower() for x in json.loads(m[db_refs]).values())
                        ]
                        # if unique match
                        if len(db_ref_matched_rows) == 1:
                            gene.load_attributes(db_ref_matched_rows[0])
                            continue


class GeneInfo(dict):
    def __init__(self, tax_id: str):
        """ Loads genes for given organism in a dict.

        Each instance of :class:`Gene` is mapped to corresponding Entrez ID

        Parameters
        ----------
        tax_id: str
            Taxonomy id of target organism.

        """
        super().__init__()
        self.tax_id: str = tax_id
        self.gene_db_path: str = self._gene_db_path()

        connection = sqlite3.connect(self.gene_db_path)
        cursor = connection.cursor()

        for gene_info in cursor.execute('SELECT * FROM gene_info').fetchall():
            gene = Gene()
            gene.load_attributes(gene_info)
            self[gene.gene_id] = gene

        cursor.close()
        connection.close()

    def _gene_db_path(self):
        return serverfiles.localpath_download(DOMAIN, f'{self.tax_id}.sqlite')


def load_gene_summary(tax_d: str, genes: List[Optional[str]]) -> List[Optional[Gene]]:
    gene_db_path = serverfiles.localpath_download(DOMAIN, f'{tax_d}.sqlite')

    # filter NoneTypes
    _genes = [g for g in genes if g]

    with contextlib.closing(sqlite3.connect(gene_db_path)) as con:
        with con as cur:

            gene_map: Dict[str, Gene] = {}
            for gene_info in cur.execute(f'SELECT * FROM gene_info WHERE gene_id in ({",".join(_genes)})').fetchall():
                gene = Gene()
                gene.load_attributes(gene_info)
                gene_map[gene.gene_id] = gene

            return [gene_map.get(gid, None) if gid else None for gid in genes]


if __name__ == "__main__":
    gm = GeneMatcher('9606')
    gm.genes = ['CD4', '614535', 'ENSG00000205426', "2'-PDE", 'HB-1Y']
    print(list(zip(gm.genes, [g.input_identifier for g in gm.genes])))
    _homologs = load_gene_summary('10090', [g.homolog_gene(taxonomy_id='10090') for g in gm.genes])
    print(_homologs)



""" OWMarkerGenes """
import sys
from typing import Set
from functools import partial
from collections import defaultdict

import numpy as np

from AnyQt.QtGui import QFont, QColor
from AnyQt.QtCore import Qt, QSize, QTimer, QModelIndex, QItemSelection, QItemSelectionModel, QItemSelectionRange
from AnyQt.QtWidgets import QLineEdit, QTreeView

from Orange.data import MISSING_VALUES, Table, Domain
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils.itemmodels import TableModel

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID, GENE_ID_COLUMN, GENE_AS_ATTRIBUTE_NAME

serverfiles_domain = 'marker_genes'


def get_available_db_sources():
    found_sources = {}

    try:
        found_sources.update(serverfiles.ServerFiles().allinfo(serverfiles_domain))
    except ConnectionError:
        raise ConnectionError('Can not connect to {}. Using only local files.'.format(serverfiles.server_url))
    finally:
        found_sources.update(serverfiles.allinfo(serverfiles_domain))
        return {item.get('title').split(': ')[-1]: item for item in found_sources.values()}


# todo: this is ugly, refactor this in the future.
class HeaderIndex:
    NAME = 0
    GENE = 1
    CELL_TYPE = 2
    FUNCTION = 3
    REFERENCE = 4
    URL = 5


HeaderLabels = {HeaderIndex.GENE: 'Entrez ID', HeaderIndex.REFERENCE: 'Reference', HeaderIndex.URL: 'URL'}


class SearchableTableModel(TableModel):

    ClassVar, Meta, Attribute = range(3)

    ColorForRole = {ClassVar: None, Meta: None, Attribute: None}

    def __init__(self, data, parent):
        TableModel.__init__(self, data, parent)
        self._data = data
        self._roleData = {Qt.DisplayRole: self.source}
        self._roleData = partial(defaultdict, partial(defaultdict, partial(defaultdict, lambda: None)))(self._roleData)
        self.set_column_links()

    def set_column_links(self):
        domain = self._data.domain
        ref_col = domain.metas.index(domain[HeaderLabels[HeaderIndex.REFERENCE]])

        font = QFont()
        font.setUnderline(True)
        color = QColor(Qt.blue)
        for i, row in enumerate(self.source):
            link = row[HeaderLabels[HeaderIndex.URL]].value
            if len(link) and 'http' in link:
                self._roleData[gui.LinkRole][i][ref_col] = link
                self._roleData[Qt.FontRole][i][ref_col] = font
                self._roleData[Qt.ForegroundRole][i][ref_col] = color

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            cell_data = super().data(index, role)
            return "" if cell_data in MISSING_VALUES else cell_data
        elif role in (gui.LinkRole, Qt.FontRole, Qt.ForegroundRole):
            row, col = index.row(), index.column()
            return self._roleData[role][row][col]
        return super().data(index, role)

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self.source)

    def wrap(self, table):
        self.beginResetModel()
        self.source = table
        self._roleData.clear()
        self.resetSorting()
        self.endResetModel()

    def filter_table(self, filter_pattern):
        selection = np.full(self._data.metas.shape, True)
        for search_word in filter_pattern.split():
            match_result = (
                np.core.defchararray.find(np.char.lower(self._data.metas.astype(str)), search_word.lower()) >= 0
            )
            selection = selection & match_result
        return selection

    def update_model(self, filter_pattern=None):
        # clear cache if model changes
        self._row_instance.cache_clear()
        self.wrap(self._data[self.filter_table(filter_pattern).any(axis=1), :])
        self.set_column_links()

    def data_length(self):
        return len(self._data)


class MarkerGroupContextHandler(settings.ContextHandler):
    def __init__(self):
        super().__init__()

    def match(self, context, group, *args):
        if not context.group == group:
            return self.NO_MATCH

        return self.PERFECT_MATCH

    def new_context(self, group):
        context = super().new_context()
        context.group = group
        return context


class OWMarkerGenes(widget.OWWidget):
    name = "Marker Genes"
    icon = 'icons/OWMarkerGenes.svg'
    priority = 170

    replaces = ['orangecontrib.single_cell.widgets.owmarkergenes.OWMarkerGenes']

    class Outputs:
        genes = widget.Output("Genes", Table)

    want_main_area = True

    selected_group: str = settings.Setting('')
    selected_db_source: str = settings.Setting('')
    filter_text: str = settings.Setting('')
    header_state: bytes = settings.Setting(b'')
    auto_commit = settings.Setting(True)

    settingsHandler = MarkerGroupContextHandler()
    selected_genes: Set[tuple] = settings.ContextSetting(set())

    def __init__(self):
        super().__init__()
        self._data = None
        self._available_db_sources = None
        self.output = None

        self._timer = QTimer()
        self._timer.timeout.connect(self._filter_table)
        self._timer.setSingleShot(True)
        self.info.set_input_summary("0")
        self.info.set_output_summary("0")

        box = gui.widgetBox(self.controlArea, 'Database', margin=0)
        self.db_source_index = -1
        self.db_source_cb = gui.comboBox(box, self, 'db_source_index')
        self.db_source_cb.activated[int].connect(self.handle_source_changed)

        box = gui.widgetBox(self.controlArea, 'Organism', margin=0)
        self.group_index = -1
        self.group_cb = gui.comboBox(box, self, 'group_index')
        self.group_cb.activated[int].connect(self.set_group_index)

        gui.rubber(self.controlArea)

        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit", "Commit Automatically")
        # TODO: to avoid this, marker genes table should have 'tax_id' column
        self.map_group_to_taxid = {'Human': '9606', 'Mouse': '10090'}

        filter_line_edit = gui.lineEdit(self.mainArea, self, "filter_text")  # type: QLineEdit
        filter_line_edit.setPlaceholderText("Filter...")
        filter_line_edit.textEdited.connect(self.call_filter_timer)

        self.view = view = QTreeView(
            rootIsDecorated=False,
            uniformRowHeights=True,
            selectionMode=QTreeView.ExtendedSelection,
            sortingEnabled=True,
        )

        view.viewport().setMouseTracking(True)
        self.mainArea.layout().addWidget(view)

        self._load_data()
        if self.header_state:
            view.header().restoreState(self.header_state)

    @property
    def available_db_sources(self) -> dict:
        return self._available_db_sources

    @available_db_sources.setter
    def available_db_sources(self, value: dict):
        self._available_db_sources = value

        items = list(value.keys())
        try:
            idx = items.index(self.selected_db_source)
        except ValueError:
            idx = -1

        self.db_source_cb.clear()
        self.db_source_cb.addItems(items)

        if idx != -1:
            self.db_source_index = idx
            self.selected_db_source = items[idx]
        elif items:
            self.db_source_index = min(max(self.db_source_index, 0), len(items) - 1)

        self.set_db_source_index(self.db_source_index)

    @property
    def data(self) -> Table:
        return self._data

    @data.setter
    def data(self, value: Table):
        """ Set the source data.

        The data is then filtered on the first meta column (group)
        """
        self._data = value
        domain = value.domain

        if domain.metas:
            group = domain.metas[0]
            groupcol, _ = value.get_column_view(group)

            if group.is_string:
                group_values = list(set(groupcol))
            elif group.is_discrete:
                group_values = group.values
            else:
                raise TypeError("Invalid column type")
            try:
                idx = group_values.index(self.selected_group)
            except ValueError:
                idx = -1

            self.group_cb.clear()
            self.group_cb.addItems(group_values)

            if idx != -1:
                self.group_index = idx
                self.selected_group = group_values[idx]
            elif group_values:
                self.group_index = min(max(self.group_index, 0), len(group_values) - 1)

            self.set_group_index(self.group_index)

    def _load_data(self):
        self.available_db_sources = get_available_db_sources()
        file_name = self.available_db_sources[self.selected_db_source]['filename']

        try:
            serverfiles.update(serverfiles_domain, file_name)
        except ConnectionError:
            raise ConnectionError('Can not connect to {}. ' 'Using only local files.'.format(serverfiles.server_url))
        finally:
            file_path = serverfiles.localpath_download(serverfiles_domain, file_name)
            data = Table(file_path)

            # enforce order
            old_domain = data.domain
            new_domain = Domain(
                [],
                metas=[
                    old_domain['Organism'],
                    old_domain['Name'],
                    old_domain['Entrez ID'],
                    old_domain['Cell Type'],
                    old_domain['Function'],
                    old_domain['Reference'],
                    old_domain['URL'],
                ],
            )
            data = data.transform(new_domain)
            self.data = data

    def set_selection(self):
        selected = self.selected_rows()

        if len(selected):
            header_count = self.view.header().count() - 1

            if self.view.model().rowCount() <= selected[-1]:
                return

            selection = QItemSelection()

            for row_index in selected:
                selection.append(
                    QItemSelectionRange(
                        self.view.model().index(row_index, 0), self.view.model().index(row_index, header_count)
                    )
                )

            self.view.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)

    def handle_source_changed(self, source_index):
        self.set_db_source_index(source_index)
        self._load_data()

    def set_db_source_index(self, source_index):
        self.closeContext()
        self.db_source_index = source_index
        self.selected_db_source = self.db_source_cb.itemText(source_index)

    def set_group_index(self, group_index):
        self.closeContext()
        self.group_index = group_index
        self.selected_group = self.group_cb.itemText(group_index)
        self._setup()

    def call_filter_timer(self, search_string):
        self._timer.stop()
        if search_string != self.filter_text:
            self.filter_text = search_string
        self._timer.start(700)

    def _filter_table(self):
        model = self.view.model()
        assert isinstance(model, SearchableTableModel)
        model.update_model(str(self.filter_text))
        self.set_selection()
        self.update_data_info()
        self.update_model()
        self.commit()

    def update_data_info(self):
        model = self.view.model()
        self.info.set_input_summary(f"Shown : {str(len(model.source))}/{str(model.data_length())}")
        self.info.set_output_summary(f"Selected: {str(len(self.selected_genes))}")

    def _setup(self):
        self.closeContext()
        data = self.data
        group = data.domain.metas[0]
        gvec = data.get_column_view(group)[0]

        if group.is_string:
            mask = gvec == self.group_cb.itemData(self.group_index, Qt.DisplayRole)
        else:
            mask = gvec == self.group_index

        data = data[mask]
        rest = data[:, data.domain.metas[1:]]
        model = SearchableTableModel(rest, parent=self)
        ref_col = rest.domain.metas.index(rest.domain[HeaderLabels[HeaderIndex.REFERENCE]])
        self.view.setItemDelegateForColumn(ref_col, gui.LinkStyledItemDelegate(self.view))

        self.view.setModel(model)
        self.view.selectionModel().selectionChanged.connect(self._on_selection_changed)

        self.openContext(self.selected_group)
        self.call_filter_timer(self.filter_text)
        self.view.hideColumn(HeaderIndex.URL)

    def _on_selection_changed(self, *args):
        self.update_model()
        self.commit()
        self.update_data_info()

    def selected_rows(self):
        """ Return row index for selected genes
        """
        model = self.view.model()

        if not self.selected_genes:
            return []

        return [
            row_index
            for row_index in range(model.rowCount())
            if (
                model.index(row_index, HeaderIndex.GENE).data(),
                model.index(row_index, HeaderIndex.CELL_TYPE).data(),
                model.index(row_index, HeaderIndex.REFERENCE).data(),
            )
            in self.selected_genes
        ]

    def commit(self):
        self.Outputs.genes.send(self.output)

    def update_model(self):
        model = self.view.model()
        assert isinstance(model, SearchableTableModel)
        rows = [mi.row() for mi in self.view.selectionModel().selectedRows(0)]

        if rows and len(rows) != len(model.source):
            rows = model.mapToSourceRows(rows)
            output = model.source[rows]
        else:
            output = model.source

        gene_id = self.view.selectionModel().selectedRows(HeaderIndex.GENE)
        cell_type = self.view.selectionModel().selectedRows(HeaderIndex.CELL_TYPE)
        ref = self.view.selectionModel().selectedRows(HeaderIndex.REFERENCE)

        self.selected_genes = {
            (entrez.data(), cell.data(), ref.data()) for entrez, cell, ref in zip(gene_id, cell_type, ref)
        }

        # always false for marker genes data tables in single cell
        output.attributes[GENE_AS_ATTRIBUTE_NAME] = False
        # set taxonomy id in data.attributes
        output.attributes[TAX_ID] = self.map_group_to_taxid.get(self.selected_group, '')
        # set column id flag
        output.attributes[GENE_ID_COLUMN] = HeaderLabels[HeaderIndex.GENE]
        output.name = 'Marker Genes'

        self.output = output

    def closeEvent(self, event):
        self.header_state = bytes(self.view.header().saveState())
        super().closeEvent(event)

    def sizeHint(self):
        return super().sizeHint().expandedTo(QSize(750, 500))


def main(argv=None):
    from AnyQt.QtWidgets import QApplication

    app = QApplication(argv or sys.argv)
    w = OWMarkerGenes()
    w.show()
    w.activateWindow()
    rv = app.exec_()
    w.saveSettings()
    w.onDeleteWidget()
    return rv


if __name__ == "__main__":
    sys.exit(main())
